#include "command.hpp"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <netinet/tcp.h> // TCP_NODELAY
#include <iostream>
#include <sstream>
#include <strings.h> // bzero, bcopy

icfp::link::link( const std::string &host, int port )
: m_socket( socket( AF_INET, SOCK_STREAM, 0 ) ),
    m_connection() {
    if ( m_socket < 0 )
        throw std::runtime_error( "Could not open socket" );
    {
        int flag = 1;
        if ( setsockopt( m_socket, IPPROTO_TCP, TCP_NODELAY, &flag, sizeof flag ) < 0 )
            throw std::runtime_error( "Could not turn off socket buffering" );
    }
    
    hostent *server = gethostbyname( host.c_str() );
    if ( !server )
        throw std::runtime_error( "Could not find host name address" );
    
    sockaddr_in address;
    bzero( &address, sizeof address );
    address.sin_family = AF_INET;
    address.sin_port = htons( port );
    bcopy( server->h_addr, &address.sin_addr.s_addr, server->h_length);

    if ( connect( m_socket, reinterpret_cast< sockaddr * >( &address ), sizeof address ) < 0 )
        throw std::runtime_error( "Could not connect" );
}

void icfp::link::send( const std::string &command ) const {
    if ( write( m_socket, command.c_str(), command.length() ) < 0 )
        throw std::runtime_error( "Couldn't write to socket" );
}

std::string icfp::link::receive() const {
    std::string line;
    char ch;
    do {
        read( m_socket, &ch, sizeof ch );
        line += ch;
    } while ( ch != ';' );
    return line;
}


bool icfp::process( const icfp::link &link, icfp::map &map, icfp::rover &rover ) {
    std::string message = link.receive();
    std::stringstream parse( message );

    char instruction;
    parse >> instruction;
    switch ( instruction ) {
        case 'E':
            return false;
        case 'I':
            parse >> map.width >> map.height;
            break;
        case 'T': {
            int time;
            parse >> time >> rover.accelerator >> rover.steering >> rover.position.first >> rover.position.second >> rover.dir >> rover.speed;

            map.obstacles.clear();
            
            char object_kind;
            do {
                parse >> object_kind;
                switch ( object_kind ) {
                    case ';':
                        break;
                    case 'h': {
                        icfp::obstacle home( parse );
                        break;
                    }
                    case 'b':
                    case 'c': {
                        icfp::obstacle obstacle( parse );
                        map.obstacles.push_back( obstacle );
                        break;
                    }
                    case 'm': {
                        icfp::position pos;
                        double dir, speed;
                        parse >> pos.first >> pos.second >> dir >> speed;
                        break;
                    }
                    default:
                        std::cout << object_kind;
                }
            } while ( object_kind != ';' );
            break;
        }
        default:
            std::cout << message << std::endl;
    }
    return true;
}
